"""Functions for dealing with example case study and synthetic data."""
import itertools
import logging
import math
import random
import re
from importlib.resources import files
from pathlib import Path
from typing import NamedTuple, Optional, Dict, Any, List, Union, Set, Callable

import numpy as np
import numpy.random
import pandas as pd

logger: logging.Logger = logging.getLogger(__name__)

class ExampleData(NamedTuple):
    """Example data, including citations, description, and other metadata."""
    data: pd.DataFrame
    """Table containing the data."""
    rank: Optional[Union[int, List[int]]]
    """Correct rank for this data, or ranks if more than one can be 
    considered correct."""
    row_metadata: Optional[pd.DataFrame]
    """Metadata associated to each row."""
    col_metadata: Optional[pd.DataFrame]
    """Metadata associated to each column."""
    other_metadata: Optional[Dict[str, Any]]
    """Any other metadata related to this data (data dictionaries etc.)"""
    description: str
    """Longform description of the data."""
    name: str
    """Descriptive name."""
    short_name: str
    """Short name."""
    doi: Optional[str]
    """DOI for data or original paper."""
    citation: Optional[str]
    """Preferred citation if you use this data."""

    def _repr_html_(self) -> str:
        """HTML output summarising dataset for display in notebooks etc."""
        html: str = (
            "<table style='font-family: monospace'>"
            "<tr><td colspan='2' style='font-weight: bold'>"
            f"{self.name}</td></tr>"
            f"<tr><td style='font-size: smaller' colspan='2'>cvanmf example "
            f"data</td></tr>\n"
            f"<tr><td>shape</td><td>{str(self.data.shape)}</td></tr>\n"
            f"<tr><td>rank</td><td>{str(self.rank)}</td></tr>\n"
            f"<tr><td>description</td><td>"
            f"{_shorten_docstring(self.description)}</td></tr>\n"
        )
        if self.doi is not None:
            html += f"<tr><td>doi</td><td>{self.doi}</td></tr>\n"
        if self.citation is not None:
            html += f"<tr><td>citation</td><td>{self.citation}</td></tr>\n"
        html += "</table>"
        return html

    def __repr__(self) -> str:
        repr: str = (
            f"ExampleData[{self.short_name}, shape={self.data.shape}, "
            f"rank={self.rank}]"
        )
        return repr


def leukemia() -> ExampleData:
    """Gene expression data for ALL and AML B- and T-cell type leukemia.

    This data was analysed in Brunet et al (2004), and often used as a
    standard dataset for biological applications of NMF since. It has two
    broad categories (ALL/AML), but AML can be refined into two subtypes (B/T).
    B-cell AML appears to contain a further stable sub-grouping, so we have
    indicated the true rank of this data as 3 or 4.
    """

    data: pd.DataFrame = pd.read_csv(
            str(files("cvanmf.data").joinpath(
                Path("ALL_AML", "ALL_AML_counts.tsv.gz")
            )),
            sep="\t",
            index_col=0
        )
    col_md: pd.DataFrame = pd.DataFrame(
        data.columns.map(__label_all_leuk),
        index=data.columns
    )
    return ExampleData(
        data=data,
        row_metadata=pd.read_csv(
            str(files("cvanmf.data").joinpath(
                Path("ALL_AML", "ALL_AML_feature_md.tsv.gz")
            )),
            sep="\t",
            index_col=0
        ),
        col_metadata=col_md,
        other_metadata=None,
        rank=4,
        description=_shorten_docstring(leukemia.__doc__),
        name="Leukemia (ALL/AML) Gene Expression",
        doi="doi:10.1073/pnas.030853110",
        citation=("Brunet, J-P., Tamayo, P., Golub, T. R. & Mesirov, "
                  "J. P. Metagenes and molecular pattern discovery using "
                  "matrix factorization. Proc. Natl. Acad. Sci. U.S.A. 101, "
                  "4164–4169 (2004)."),
        short_name="ALL_AML"
    )


def swimmer() -> ExampleData:
    """Stick figure images of a swimmer with 4 limbs in 4 positions.

    Designed by Donoho et al. to be partially representable by NMF, each image
    has a line torso, accompanied by four straight limbs which can be in one
    of four positions (0, 45, 90, 135 and 180 degrees from torso). With the
    exception of the torso, each limb position should be representable by
    NMF decomposition. As such, the true rank of this data is 17 (4*4 limbs,
    plus torso), but with conventional NMF as implemented here the torso cannot
    be learnt, only the 16 limbs.
    """

    return ExampleData(
        data=pd.read_csv(
            str(files("cvanmf.data").joinpath(
                Path("swimmer", "swimmer.tsv.gz")
            )),
            sep="\t",
            index_col=0
        ),
        rank=17,
        row_metadata=None,
        col_metadata=None,
        other_metadata=None,
        description=_shorten_docstring(swimmer.__doc__),
        name="Swimmer Dataset",
        doi="",
        citation=("Donoho, D. & Stodden, V. When Does Non-Negative Matrix "
                  "Factorization Give a Correct Decomposition into Parts? in "
                  "Advances in Neural Information Processing Systems (eds. "
                  "Thrun, S., Saul, L. & Schölkopf, B.) vol. 16 (MIT Press, "
                  "2003)."),
        short_name="swimmer"
    )


def lung_cancer_cells() -> ExampleData:
    """Relative cell-compositions from non-small cell lung cancer studies.

    Gives the number of cells of different types in lung tissue samples from a
    non-small cell lung cancer atlas, which was compiled from 29 studies and
    includes 556 samples, from 318 individuals (86 of which are healthy
    controls). The data was uploaded to cellxgene using their standard
    ontologies, which is the source we have taken the data from. Metadata
    provided here is a mixture of metadata from cellxgene, and some from the
    original paper. We have selected out only the tissues samples labelled as
    "lung". In total, this gives 224 samples, and 33 cell types. The data
    here is total-sum-scaled, i.e. each sample sums to 1.
    """

    return ExampleData(
        data=pd.read_csv(
            str(files("cvanmf.data").joinpath(
                Path("NSCLC", "nsclc.lung.composition.tsv.gz")
            )),
            sep="\t",
            index_col=0
        ),
        rank=[],
        row_metadata=None,
        col_metadata=pd.read_csv(
            str(files("cvanmf.data").joinpath(
                Path("NSCLC", "nsclc.lung.metadata.tsv.gz")
            )),
            sep="\t",
            index_col=0
        ),
        other_metadata=None,
        description=_shorten_docstring(lung_cancer_cells.__doc__),
        name="Non-small cell lung cancer",
        doi="https://doi.org/10.1016/j.ccell.2022.10.008",
        citation=("Salcher, S. et al. High-resolution single-cell atlas "
                  "reveals diversity and plasticity of tissue-resident "
                  "neutrophils in non-small cell lung cancer. Cancer Cell 40, "
                  "1503-1520.e8 (2022). \n"
                  "Program, C. S.-C. B. et al. CZ CELL×GENE Discover: A "
                  "single-cell data platform for scalable exploration, "
                  "analysis and modeling of aggregated data. "
                  "2023.10.30.563174 Preprint at "
                  "https://doi.org/10.1101/2023.10.30.563174 (2023)."),
        short_name="NSCLC"
    )


def synthetic_blocks(m: int = 100,
                     n: int = 100,
                     overlap: float = 0.25,
                     k: int = 3,
                     normal_noise_params: Optional[Dict] = None,
                     scale_lognormal_params: Optional[Dict] = None
                     ) -> ExampleData:
    """Generate simple synthetic data.

    Create an m x n matrix with blocks along the diagonal which overlap to an
    extent defined by overlap.

    :param m: Number of rows in matrix
    :param n: Number of columns in matrix
    :param overlap: Proportion of block length to participate in overlap
    :param k: Number of signatures
    :param normal_noise_params: Parameters to pass to `numpy.random.normal`
        to apply noise to entries. Leave as none to use default parameters.
    :param scale_lognormal_params: Parameters to pass to
        `numpy.random.lognormal` to scale each feature (give some features
        higher values than others). If set to true, will use default parameters
        for distribution. Leave as None to skip feature scaling.
    """

    # Matrix dimensions
    i, j = m, n

    # Noise parameters
    do_scale: bool = scale_lognormal_params is not None
    scale_lognormal_params: Dict = (
        dict() if scale_lognormal_params is None or
        not isinstance(scale_lognormal_params, dict)
        else scale_lognormal_params
    )
    normal_noise_params = (
        dict() if normal_noise_params is None
             else normal_noise_params
    )

    # Width of blocks without overlap
    base_h, tail_h = divmod(i, k)
    base_w, tail_w = divmod(j, k)
    # Overlap proportion - proportion of block's base dimension to extend
    # block by
    overlap_proportion: float = overlap
    overlap_h: int = math.ceil(base_h * overlap_proportion)
    overlap_w: int = math.ceil(base_w * overlap_proportion)
    # Make a randomly filled matrix, multiply by mask matrix which has 0
    # or 1 then apply noise (so 0s also have some noise)
    mask: np.ndarray = np.zeros(shape=(i, j))
    for ki in range(k):
        h: int = base_h + tail_h if k == ki else base_h
        w: int = base_w + tail_w if k == ki else base_w
        h_start: int = max(0, ki * base_h)
        h_end: int = min(i, h_start + base_h + overlap_h)
        w_start = max(0, ki * base_w)
        w_end: int = min(j, w_start + base_w + overlap_w)
        mask[h_start:h_end, w_start:w_end] = 1.0
    block_mat: np.ndarray = np.random.uniform(
        low=0.0, high=1.0, size=(i, j)) * mask
    # Apply noise from normal distribution
    normal_noise_params = dict(loc=0.0, scale=0.1) | normal_noise_params
    block_mat = block_mat + np.absolute(
        np.random.normal(size=(i, j), **normal_noise_params))
    # Scale features
    if do_scale:
        fscale: np.ndarray = np.random.lognormal(
            **scale_lognormal_params,
            size=m
        )
        block_mat = fscale.reshape(-1, 1) * block_mat
    # Convert to DataFrame and add some proper naming for rows/cols
    df: pd.DataFrame = pd.DataFrame(
        block_mat,
        index=[f'feat_{i_lab}' for i_lab in range(i)],
        columns=[f'samp_{j_lab}' for j_lab in range(j)])
    # TSS scale
    df = df / df.sum()
    return ExampleData(
        data=df,
        rank=k,
        row_metadata=None,
        col_metadata=None,
        other_metadata=dict(
            overlap=overlap,
            scale_lognormal_params=scale_lognormal_params,
            normal_noise_params=normal_noise_params
        ),
        doi="",
        citation="",
        description="Synthetic data with blocks along the diagonal which "
                    "overlap to a defined extent.",
        name="Synthetic Overlapping Blocks",
        short_name="SYNTH_BLOCKS",
    )


def synthetic_dense(m: int = 100,
                    n: int = 100,
                    h_sparsity: float = 0.0,
                    shared_features: float = 0.25,
                    k: int = 3,
                    normal_noise_params: Optional[Dict] = None,
                    scale_lognormal_params: Optional[Dict] = None,
                    keep_mats: bool = False
                    ) -> ExampleData:
    """Generate dense synthetic data.

    Dense data is generated by making a :math:`W` matrix with :math:`k`
    signatures, and multiplying this with a randomly filled :math:`H` matrix.
    Optionally, a proportion of the :math:`H` matrices can be randomly set to 0
    using `h_sparsity`. The extent to which features are shared between
    signatures is defined via `shared_features`. Each signature is initially
    assigned an even proportion of the :math:`m` features (remainder spread as
    evenly as possible between them), so there are no shared features. Then if
    :math:`|k|` is the number of features assigned to a signature, each
    signature is assigned :math:`|k|*shared\_features` randomly selected from
    the remaining features. This means the overlapping structure is potentially
    quite different from that of :func:`synthetic_blocks`.

    :param m: Number of features.
    :param n: Number of samples.
    :param h_sparsity: Proportion of :math:`H` matrix to randomly set to 0
    :param shared_features: Amount of shared features to add to a signature, as
        a proportion of it's base size.
    :param k: Number of signatures.
    :param normal_noise_params: Parameters passed to
        :func:`numpy.random.normal` when adding noise to data.
    :param scale_lognormal_params: Parameters passed to
        :func:`numpy.random.lognormal` when selecting weights for features in a
        signature. If this is None, a uniform distribution between 0 and 1 is
        used instead.
    :param keep_mats: Return the :math:`H` and :math:`W` matrices used to
        generate the data in the :attr:`ExampleData.other_metadata`.
    """

    # Create W
    k_scale: np.array = np.array([1] * k)
    k_feats: np.array = np.floor((k_scale / k_scale.sum()) * m)
    rem: int = m - int(k_feats.sum())
    # Got to be a nicer way to do this
    for i in range(rem):
        k_feats[i] = k_feats[i] + 1
    k_feat_names: List[Set[str]] = [
        set(f'sig{i}_feat{j}' for j in range(int(k_feats[i]))) for i in range(k)
    ]
    # Add shared features
    all_feats: Set[str] = set(itertools.chain.from_iterable(k_feat_names))
    for k_names in k_feat_names:
        other: List[str] = random.choices(
            list(all_feats.difference(k_names)),
            k=math.floor(len(k_names) * shared_features)
        )
        k_names.update(other)
    w: np.ndarray
    if scale_lognormal_params:
        w = np.random.lognormal(size=(m, k), **scale_lognormal_params)
    else:
        w = np.random.uniform(size=(m, k), low=0.0, high=1.0)
    mask: pd.DataFrame = pd.DataFrame(
        np.zeros(shape=(m, k)),
        index=list(all_feats),
        columns=[i for i in range(k)]
    ).sort_index()
    mask = mask.apply(lambda x: x.index.isin(k_feat_names[x.name]))
    w = w * mask.values

    # Create H
    h: np.ndarray = np.random.uniform(low=0.0, high=1.0, size=n*k)
    h[numpy.random.choice(
        list(range(n*k)),
        size=math.floor(n*k*h_sparsity),
        replace=False
    )] = 0.0
    h = h.reshape(k, n)

    wh: pd.DataFrame = pd.DataFrame(
        w.dot(h),
        index=mask.index,
        columns=[f's{i}' for i in range(n)]
    )
    normal_noise_params = (dict() if normal_noise_params is None else
                           normal_noise_params)
    normal_noise_params = dict(loc=0.0, scale=0.1) | normal_noise_params
    wh = (wh + numpy.random.normal(size=wh.shape,
                                   **normal_noise_params)).clip(lower=0)

    other_metadata: Dict[str, Any] = dict(
        h_sparsity=h_sparsity,
        shared_features=shared_features,
        scale_lognormal_params=scale_lognormal_params,
        normal_noise_params=normal_noise_params
    )
    if keep_mats:
        other_metadata['h'] = h
        other_metadata['w'] = w

    return ExampleData(
        data=wh,
        col_metadata=None,
        row_metadata=mask,
        rank=k,
        other_metadata=other_metadata,
        name="Synthetic Dense Data",
        short_name="SYNTH_DENSE",
        doi="",
        citation="",
        description=""
    )


def example_abundance() -> pd.DataFrame:
    """Genus level Non-Western cohort bacterial microbiome abundance.

    From Frioux et al. (2023, https://doi.org/10.1016/j.chom.2023.05.024).

    :return: Genus level relative abundance table using GTDB r207 taxonomy.
    :rtype: pd.DataFrame
    """
    return pd.read_csv(
        str(files("cvanmf.data").joinpath("NW_ABUNDANCE.tsv")),
        sep="\t",
        index_col=0
    )

def _shorten_docstring(s: str) -> str:
    """Strip newlines and multiple spaces from a docstring to make it a more
    readable format when returned."""
    # Want to keep any double newlines, but remove any singles
    # This is a very messy way of doing it but didn't have time to work out a
    # regex
    s = (
        s.replace("\n\n", ":lb:")
        .replace("\n", "")
        .replace(":lb:", "\n\n")
    )
    s = re.sub(r"[\t ]+", " ", s)
    return s

def __label_all_leuk(smpl: str) -> str:
    if smpl[:3] == "AML":
        return "AML"
    if "B-cell" in smpl:
        return "ALL B-Cell"
    if "T-cell" in smpl:
        return "ALL T-Cell"
    raise ValueError("Leukemia sample with incorrect label encountered")






"""Example data for decomposition.

This modules provides both real world datasets, and functions to synthetic data
with a known number of signatures. Each is returned as an
:class:`ExampleData` object, which includes metadata and
citations for the data where appropriate. Each can be loaded by calling the
relevant function (i.e. `data.leukemia()`)

Real Data
---------
:func:`leukemia` is gene expression data from ALL/AML-type leukemia
patients. :func:`lung_cancer_cells` is the count of types of cells in
non-small cell lung cancer tissues. :func:`swimmer` is an image dataset
with stick figure representations of a swimmer from above.

Synthetic Data
--------------
:func:`synthetic_blocks` makes data with an overlapping block pattern along
the diagonal. :func:`synthetic_dense` makes data with a dense structure
(each sample can contain any number of signatures).
"""

from cvanmf.data.utils import (leukemia, swimmer,
                               lung_cancer_cells, example_abundance,
                               synthetic_blocks, synthetic_dense, ExampleData)



#!/usr/bin/env python
"""
Reapply existing Enterosignature models to new abundance data.

The easiest way to do this is through the :func:`reapply` function, which is
most flexible about parameter types. The other functions perform individual
steps, which are useful if you want fine control of a given step, but
probably not necessary for most uses.
"""
import logging
import pathlib
import re
from typing import (Collection, Dict, Iterator, List,
                    Optional, Set, Tuple, Union, Protocol)

import click
import numpy as np
import pandas as pd
from sklearn.decomposition import non_negative_factorization

from cvanmf.models import Signatures

logger: logging.Logger = logging.getLogger(__name__)

# Compile regular expressions
RE_RANK: re.Pattern = re.compile("[a-zA-Z]__")
RE_SPLIT_GENUS: re.Pattern = re.compile(r"([\w]+)_\w$")
# Match the last named rank, plus any following ?s
RE_SHORTEN: re.Pattern = re.compile(r".*;([^\;?]+[;\?]*)$")


class CVANMFException(Exception):
    """cvanmf specific exception."""


class FeatureMapping:
    """Connect new data table features to those in the model.

    Manage the mappings from input features to model features. Source features
    are the features in the new abundance table we want to fit to the model;
    target features are the features in the model we're trying to match to.
    User defined mappings can be provided via `hard_map`, any subsequent
    mappings for a source taxon in `hard_map` will be ignored. New mappings are
    added via :meth:`add`. When mappings are fully defined the model w matrix
    and the new data table can be matched using :meth:`transform_w` and
    :meth:`transform_abundance`
    """

    def __init__(self,
                 target_features: Set[str],
                 source_features: Set[str],
                 hard_map: Optional[Dict[str, str]] = None):
        """
        :param target_features: Model features to map to
        :type target_features: Set[str]
        :param source_features: Input features to be mapped from
        :type source_features: Set[str]
        :param hard_map: User defined mappings, as a dictionary with source
            as key and target as value.
        :type hard_map: Optional[Dict[str, str]]
        """

        # hard_map are user provided mappings which will never be altered
        self.__target_features = target_features
        self.__source_features = source_features
        self.__hard_map = {} if hard_map is None else hard_map
        self.__map: Dict[str, List[str]] = dict()

    def add(self, feature_from: str, feature_to: str) -> None:
        """Add a mapping. If there is already a mapping from this feature, we
        will append this one. Use :meth:`conflicts` to identify where more than
        one mapping exists.

        :param feature_from: Feature in the new table
        :type feature_from: str
        :param feature_to: Model feature to map to
        :type feature_to: str
        :raises EnteroException: Feature not in the relevant sets
        """
        # If user has provided a hard mapping, ignore this mapping
        if feature_from in self.__hard_map:
            return
        # If the genus being mapped to is not valid, raise an exception
        if feature_to not in self.__target_features:
            raise CVANMFException(
                f"Mapping of '{feature_from}' to target '{feature_to}' not "
                f"possible as target does not exist.")
        if feature_from not in self.__source_features:
            raise CVANMFException(
                f"Mapping of '{feature_from}' to target '{feature_to}' not "
                f"possible as source does not exist.")
        if feature_from not in self.__map:
            self.__map[feature_from] = [feature_to]
        else:
            self.__map[feature_from].append(feature_to)

    def to_df(self) -> pd.DataFrame:
        """Produce a dataframe of the mapping. Where mappings are amibiguous,
        multiple rows will be included. Where mappings are missing, one row with
        a blank target will be included.

        :return: DataFrame with two columns, first source feature, second target
            feature.
        :rtype: pd.DataFrame
        """

        # Could be more nicely done with some maps, but take a simple approach
        df_src: List[Tuple[str, str]] = []
        for source, targets in self.mapping.items():
            for target in targets:
                df_src.append((source, target))
        missing: List[Tuple[str, str]] = list(
            zip(self.missing(), [''] * len(self.missing())))
        df_src = df_src + missing
        df: pd.DataFrame = pd.DataFrame(
            df_src, columns=['input_feature', 'model_feature'])
        return df

    def missing(self) -> Collection[str]:
        """Identify input features which currently have no mapping.

        :return: Source features which are not mapping to any model feature
        :rtype: Collection[str]
        """
        return self.__source_features.difference(set(self.mapping.keys()))

    def transform_abundance(self, abd_tbl: pd.DataFrame) -> pd.DataFrame:
        """Applying mapping to the input table.

        Make a table with renamed and combined rows based on the
        identified mappings.

        :param abd_tbl: New table, samples on columns
        :type abd_tbl: pd.DataFrame
        :return: Table with mappings applied
        :rtype: pd.DataFrame
        """

        # Probably we could do some smarter changes and groupbys
        # TODO(apduncan): Refactor, slow and ugly
        abd: pd.DataFrame = abd_tbl.copy()
        for input_feature, feature_maps in self.mapping.items():
            # Where there are conflicting mappings, use the first one
            feature_map = feature_maps[0]
            if input_feature == feature_map:
                continue
            # Add or sum
            if feature_map not in abd.index:
                abd.loc[feature_map] = abd.loc[input_feature]
            else:
                abd.loc[feature_map] += abd.loc[input_feature]
            abd = abd.drop(labels=[input_feature])
        return abd

    def transform_w(self,
                    w: pd.DataFrame,
                    abd_tbl: pd.DataFrame) -> pd.DataFrame:
        """Match the model w matrix to the new table.

        Make a W matrix which has features not in the abundance table removed,
        and rows added for features which are in the abundance table but not the
        model.

        :param w: Model W matrix
        :type w: pd.DataFrame
        :param abd_tbl: New matrix. Should `not` have been transformed
            with :meth:`transform_abundance`.
        :type abd_tbl: pd.DataFrame
        :return: W matrix matched to new table
        :rtype: pd.DataFrame
        """
        w_new: pd.DataFrame = w.copy()
        # Drop taxa which are not in the input matrix from the W matrix
        w_feat: Set[str] = set(w_new.index)
        w_new = w_new.drop(labels=w_feat.difference(set(abd_tbl.index)))
        # Add taxa which are in the input but not in the W matrix
        missing_features: Set[str] = set(abd_tbl.index).difference(w_feat)
        missing_df: pd.DataFrame = pd.DataFrame(
            np.zeros(shape=[len(missing_features), w.shape[1]]),
            index=list(missing_features),
            columns=w.columns
        )
        w_new = pd.concat([w_new, missing_df])
        return w_new

    @property
    def conflicts(self) -> List[Tuple[str, List[str]]]:
        """Features for which more than one target exists."""
        return list(filter(lambda x: len(x[1]) > 1, self.mapping.items()))

    @property
    def mapping(self) -> Dict[str, List[str]]:
        """Mapping from source to target features."""
        return {**self.__map, **{x: [y] for x, y in self.__hard_map.items()}}


def _is_taxon_unknown(taxon: str) -> bool:
    # Remove ?, if all entries in lineage blank, the is unknown
    return all(
        map(
            lambda x: len(x.strip()) == 0,
            taxon.replace("?", "").split(';')
        )
    )


def _contain_rank_indicators(taxon: str) -> bool:
    # Check if this taxonomy contains rank indicators
    return any(map(
        lambda x: RE_RANK.search(x.strip()) is not None,
        taxon.split(';')
    ))


def _shorten_genus(
        genus: str,
        genera_list: Collection[str],
        short_genera_list: Collection[str]
) -> str:
    # Copying the operation of Clemence's shorten_genus function
    # Get last named rank, and subsequent ?
    # Genera list should the genera list, but with RE_SHORTEN applied, not 
    # applying each call as computationally inefficient (especially if we want 
    # to run on streamlit's servers)
    # Count occurences in the genera list
    short: str = RE_SHORTEN.sub(r"\1", genus)
    nb_occ: int = sum(map(lambda x: short in x, genera_list))
    nb_occ_s: int = sum(map(lambda x: short in x, short_genera_list))
    # shorten2 is the last three bits of the lineage, if we observed short in 
    # the short form genera list more than once, else we just keep short
    shorten2: str = ";".join(genus.split(';')[-3:]) if nb_occ_s > 1 else short
    final_shorten: str = short
    if nb_occ != 1:
        if nb_occ_s == 1:
            # Extract the final 2
            final_shorten = ";".join(genus.split(";")[-2:])
        else:
            final_shorten = shorten2
    return final_shorten


def _final_rank(taxon: str) -> str:
    # Exctract final non-blank or ? rank
    return next(filter(
        lambda x: len(x.strip()) != 0,
        reversed(taxon.replace("?", "").split(';'))
    ))


def _final_rank_equal(a: str, b: str) -> bool:
    return _final_rank(a) == _final_rank(b)


def _pad_lineage(
        taxon: str,
        delim: str = ";",
        length: int = 6,
        unknown: str = "") -> str:
    """Pad lineage out to the expected length using delimiters. Default is to
    genus level.

    :param str taxon: Lineage string for taxon
    :param str delim: Delimiter between ranks
    :param int length: Rank to truncate to
    :param str unknown: Text to fill in unknown ranks
    """

    lineage_parts: List[str] = taxon.split(delim)
    diff: int = length - len(lineage_parts)
    if diff > 0:
        lineage_parts += [unknown] * diff
    lineage_parts = lineage_parts[0:length]
    # Replace any length 0 entries with the unknown indicator
    lineage_parts = list(map(
        lambda x: x if len(x) > 0 else unknown,
        lineage_parts
    ))
    return delim.join(lineage_parts)


def validate_genus_table(abd_tbl: pd.DataFrame,
                         **kwargs
                         ) -> pd.DataFrame:
    """Basic checks and transformations of the abundance table. 
    
    Some transformations may be made here, such as transposition. Any 
    transformation will be written out to inform the user. Transformations are 
    done in place.
    
    :param abd_tbl: Abundance table to check
    :type abd_tbl: pd.DataFrame
    :param logger: Function to report errors
    :type logger: Callable[[str], None]
    :returns: Validated, potentially transformed, dataframe
    :rtype: pd.DataFrame
    """

    # Check the dimensions make sense
    if abd_tbl.shape[0] < 2 or abd_tbl.shape[1] < 2:
        logger.critical(
            "Table has one or fewer columns or rows. Check delimiters and "
            "newline formats are correct.")
        raise CVANMFException("Table incorrect format.")

    # Check that there were column names in the file. We're going to assume
    # that if column names are all numeric values, there were no headers.
    # Unfortunate for people who only gave their samples numbers, but alas.
    all_numeric: bool = all(map(lambda x: str(x).isnumeric(), abd_tbl.columns))
    if all_numeric:
        logger.error(
            "Table appear to lack sample IDs in the first row. Add sample "
            "IDs, or ensure all sample IDs are not numeric.")
        raise CVANMFException("Table lacks sample IDs in first row.")

    # Check that taxa are on rows. We'll do that by looking for "Bacteria;"
    # in the column names.
    count_bac: int = len(
        list(filter(lambda x: "BACTERIA;" in x.upper(), abd_tbl.columns)))
    if (count_bac / len(abd_tbl.columns)) > 0.2:
        logger.warning("Table appears to have taxa on columns, so we have "
                       "transposed it.""")
        abd_tbl = abd_tbl.T

    # Remove ? from lineages, unknown should be blank
    # abd_tbl.index = list(map(lambda x: x.replace("?", ""), abd_tbl.index))

    # See if there are rank indicators in the taxa names
    # We'll take a look at the first 10 taxa, and see if they do
    rank_indicated: bool = all(map(_contain_rank_indicators,
                                   abd_tbl.index[:10]))
    if rank_indicated:
        logger.info(
            "Taxa names appear to contain rank indicators (i.e k__, p__), "
            "these have been removed to match Enterosignature format.""")
        abd_tbl.index = map(
            lambda x: re.sub(RE_RANK, "", x),
            abd_tbl.index
        )

    # Delete any taxa which have numeric labels (basically dealing with
    # MATAFILER output which has a -1 row included). More useful for our
    # group, limited use for other maybe.
    numeric_taxa: List[str] = list(filter(
        lambda x: str(x).lstrip('-').isnumeric(), abd_tbl.index))
    if len(numeric_taxa) > 0:
        logger.info(
            f"{len(numeric_taxa)} taxa had numeric labels and were dropped.")
        abd_tbl = abd_tbl.drop(labels=numeric_taxa)

    # Pad lineages to standard length with semicolons
    new_index: pd.Index = pd.Index(map(_pad_lineage, abd_tbl.index))
    # If there are duplicated rows (same genus, or malformed strings) then
    # report an error
    duplicates: np.ndarray = new_index.duplicated(keep=False)
    if any(duplicates):
        orig_dups: List[str] = abd_tbl.index[duplicates]
        new_dups: List[str] = new_index[duplicates]
        logger.error(
            f"{len(orig_dups)} taxa are duplicates after trimming to "
            "genus length. This may be genuine duplicates, or could be "
            "due to erroneous semi-colons. Duplicates are:")
        for o, n in zip(orig_dups, new_dups):
            logger.error(f'{o} -> {n}')
        raise CVANMFException("Duplicate taxa in input after lineage " +
                              "truncated.")

    # Remove any taxa which are completely unknown (all ? or blank)
    bad_taxa: List = list(filter(_is_taxon_unknown, abd_tbl.index))
    abd_tbl = abd_tbl.drop(labels=bad_taxa)
    if len(bad_taxa) > 0:
        logger.info(f"Removed {len(bad_taxa)} unknown taxa: {bad_taxa}")

    # Remove any taxa which had 0 observations, and samples for which 0
    # taxa were observed (unlikely but check)
    zero_taxa: List = list(abd_tbl.loc[abd_tbl.sum(axis=1) == 0].index)
    zero_samples: List = list(
        abd_tbl.loc[:, abd_tbl.sum(axis=0) == 0].columns)
    if len(zero_taxa) > 0:
        abd_tbl = abd_tbl.drop(labels=zero_taxa)
        logger.info(
            f"Dropped {len(zero_taxa)} taxa with no observations: {zero_taxa}")
    if len(zero_samples) > 0:
        abd_tbl = abd_tbl.drop(columns=zero_samples)
        logger.info(
            f"Dropped {len(zero_samples)} sample with no observations")

    # Renormalise (TSS)
    abd_tbl = abd_tbl / abd_tbl.sum(axis=0)

    return abd_tbl


def match_genera(
        w: pd.DataFrame,
        y: pd.DataFrame,
        hard_mapping: Optional[Dict[str, str]] = None,
        family_rollup: bool = True,
        **kwargs
) -> FeatureMapping:
    """Match taxonomic names in the input table and the Enterosignatures W 
    matrix.
    
    This function is currently based on the R script provided by Clemence in
    the Enterosignatures (ES) gitlab repo (prepare_matrices.R)
    https://gitlab.inria.fr/cfrioux/enterosignature-paper/. 
    This will attempt to match names. Mappings in the ``hard_mapping`` parameter
    are new names to ES names, and will be applied before any other matches
    identified.
    
    :param w: Enterosignatures W matrix
    :type w: pd.DataFrame
    :param y: Abundance table being transformed
    :type y: pd.DataFrame
    :param hard_mapping: Mapping from input to ES name
    :type hard_mapping: Dict[str, str]
    :param family_rollup: Move abundance of genera which are not matched to
        the family level entry if one exists
    :type family_rollup: bool
    :param logger: Function to log messages
    :type logger: Callable[[Any], None]
    :returns: Transformed abundance table, es W matrix, and mapping object
    :rtype: Tuple[pd.DataFrame, pd.DataFrame, List[str]]
    """

    # Set optional parameter values
    if hard_mapping is None:
        hard_mapping = dict()
    # Sets of taxa
    es_taxa: Set[str] = set(w.index)
    input_taxa: Set[str] = set(y.index)
    unmatched: Set[str] = set(y.index)

    # Create an object to collect all the mappings we identify
    # No tables will be altered until we've exhausted the ways in which we
    # could generate mappings
    mapping: FeatureMapping = FeatureMapping(
        hard_map=hard_mapping,
        source_features=input_taxa,
        target_features=es_taxa)

    # Add any exact matches. If there was an exact match, remove from unmatched
    # pool
    exact = es_taxa.intersection(input_taxa)
    logger.info(
        f"{len(exact)} of {len(input_taxa)} taxa names matched exactly")
    for genus in exact:
        mapping.add(genus, genus)
    unmatched = unmatched.difference(exact)

    # Merge some split genera (ie Clostridia_A -> Clostridia)
    trimmed: Set[str] = set()
    for taxon in filter(lambda x: "_" in x.split(';')[-1], unmatched):
        root: str = re.sub(RE_SPLIT_GENUS, r"\1", taxon)
        if root in es_taxa:
            trimmed.add(taxon)
            mapping.add(taxon, root)
    logger.info(
        f"{len(trimmed)} genera trimmed and matched (i.e Ruminococcus_C ->"
        " Ruminococcus)")
    unmatched = unmatched.difference(trimmed)

    # Homogenise taxa of input matrix and ES
    # This is adapted directly from Clemence's script, and identifies taxa
    # where the lowest known rank matches, but the preceding elements do not
    es_taxa_short: List[str] = list(map(
        lambda x: RE_SHORTEN.sub(r"\1", x), es_taxa))
    input_taxa = set(unmatched)
    input_taxa_short: List[str] = list(map(
        lambda x: RE_SHORTEN.sub(r"\1", x), input_taxa))
    es_only: Set[str] = es_taxa.difference(input_taxa)
    input_only: Set[str] = input_taxa.difference(es_taxa)

    es_only_genus: Set[str] = set(
        map(lambda x: _shorten_genus(
            x,
            genera_list=es_taxa,
            short_genera_list=es_taxa_short
        ),
            es_only
            )
    )
    input_only_genus: Set[str] = set(
        map(lambda x: _shorten_genus(
            x,
            genera_list=input_taxa,
            short_genera_list=input_taxa_short
        ),
            input_only
            )
    )

    # Which can be merged based on Clemence's critera
    fixable_taxa: Set[str] = es_only_genus.intersection(input_only_genus)
    clemence_merged: Set[str] = set()
    for taxon in fixable_taxa:
        t_from: List[str] = list(filter(lambda x: taxon in x, unmatched))
        t_to: List[str] = list(filter(lambda x: taxon in x, w))
        for f in t_from:
            for t in t_to:
                mapping.add(f, t)
            clemence_merged.add(f)
    logger.info(f"{len(clemence_merged)} taxa merged by final rank (e.g. "
                 "UBA1435, CAG-314)")
    unmatched = unmatched.difference(clemence_merged)

    # Final rank matching
    final_matched: Set[str] = set()
    for taxon in unmatched:
        final_match: Iterator[str] = filter(
            lambda x: _final_rank_equal(taxon, x),
            es_taxa
        )
        for match in final_match:
            mapping.add(taxon, match)
            final_matched.add(taxon)
    logger.info(
        f"{len(final_matched)} taxa matched by name of lowest rank alone")
    unmatched = unmatched.difference(final_matched)

    # Roll any genera which are unknown, but where we have a family entry,
    # up into the family entry
    if family_rollup:
        family_match: Set[str] = set()
        for taxon in unmatched:
            family: str = ";".join(taxon.split(";")[:-1]) + ";"
            if family in es_taxa:
                family_match.add(taxon)
                mapping.add(taxon, family)
        logger.info(
            f"{len(family_match)} genera rolled up to family "
            "(i.e Lachonspiraceae;CAG-95 -> Lachnospiraceae)")
        unmatched = unmatched.difference(family_match)

    # # Get updated abundance tables and W matrix
    # new_abd = mapping.transform_abundance(abd_tbl)
    # new_w = mapping.transform_w(es_w, new_abd)
    # # Match their ordering
    # new_abd = new_abd.loc[new_w.index]
    #
    # # Summarise loss of abundance and of W weights so user can assess whether
    # # this is acceptable
    # input_unique = new_w[new_w.sum(axis=1) == 0].index
    # input_abd_missed: float = new_abd.loc[input_unique].sum().sum()
    # input_abd_missed_prop: float = input_abd_missed / new_abd.sum().sum()
    # logger(f"{len(input_unique)} taxa unique to input, could not be matched" + \
    #        f" any entries in Enterosignatures matrix.")
    # logger(f"These unique taxa represent {input_abd_missed:.2f} abundance of" + \
    #        f" a total {new_abd.sum().sum():.2f} ({input_abd_missed_prop:.2%}).")
    #
    # # Loss of W weight
    # w_total: float = es_w.sum().sum()
    # new_w_total: float = new_w.sum().sum()
    # logger(f"Sum of weights in new W matrix is {new_w_total:.2f}; " + \
    #        f"Original Enterosignatures matrix sum is {w_total:.2f}. " + \
    #        f"{1 - (new_w_total / w_total):.2%} of ES matrix weight lost due to " + \
    #        "taxa mismatch.")
    return mapping


def match_identical(w: pd.DataFrame, y: pd.DataFrame, **kwargs) -> (
        FeatureMapping):
    """Match features by identical labels only.

    :param w: W matrix from model
    :param y: Table of new data
    """
    target, source = set(w.index), set(y.index)
    mapping: FeatureMapping = FeatureMapping(
        target_features=target, source_features=source
    )
    for f in target.intersection(source):
        mapping.add(f, f)
    return mapping


def nmf_transform(new_abd: pd.DataFrame,
                  w_prime: pd.DataFrame,
                  ) -> pd.DataFrame:
    """Transform the input data into model weights.
    
    Takes the matched up W matrix and feature matrix. Expects the row
    ordering of W and feature matrix to be the same. Any NA values will be
    filled with 0.

    :param new_abd: Feature matrix matched to W
    :param w_prime: Model weights
    :return: Model weights for the given model and abundances, note this is
        not relative abundance (do not sum to 1)
    """
    # TODO: Parameters for NNLS?
    h, _, _ = non_negative_factorization(
        new_abd.T.values,
        n_components=w_prime.shape[1],
        init="custom",  # Must be custom as providing H init
        verbose=False,
        solver="mu",
        max_iter=2000,
        random_state=None,
        l1_ratio=0.0,
        beta_loss="kullback-leibler",
        update_H=False,
        H=w_prime.T.values
    )
    h_df: pd.DataFrame = pd.DataFrame(
        h,
        index=new_abd.columns,
        columns=w_prime.columns
    )
    return h_df


class FeatureMatch(Protocol):
    """Signature for functions which perform feature matching."""

    def __call__(self,
                 w: pd.DataFrame,
                 y: pd.DataFrame,
                 **kwargs) -> FeatureMapping:
        ...


class InputValidation(Protocol):
    """Signature for functions which perform input validation."""

    def __call__(self,
                 y: pd.DataFrame,
                 **kwargs) -> pd.DataFrame:
        ...


def _reapply_model(
        y: pd.DataFrame,
        w: pd.DataFrame,
        colors: Optional[List[str]],
        input_validation: InputValidation,
        feature_match: FeatureMatch,
        **kwargs
) -> 'denovo.Decomposition':
    """Reapply a model to new data Y.

    New observations using the same features can be transformed to give
    signature weights for the new observations. However, for many types of
    data is unlikely the features will be in the exact same format, and
    that the exact same types of features will be observed.

    This function uses custom functions provided to validate and transform
    input, and then to match features. Note these are applied in this order if
    you apply any transformations such a log transform in validation this may
    have undesirable effects if features are combined and summed during
    feature matching.
    """
    from cvanmf.denovo import Decomposition, NMFParameters

    new_abd: pd.DataFrame = input_validation(y, **kwargs)
    mapping: FeatureMapping = feature_match(w, new_abd, **kwargs)

    # Make W' and Y' with matching row labels
    # Get updated abundance tables and W matrix
    new_abd = mapping.transform_abundance(new_abd)
    new_w = mapping.transform_w(w, new_abd)
    # Match their ordering
    new_abd = new_abd.loc[new_w.index]

    # Summarise loss of abundance and of W weights so user can assess whether
    # this is acceptable
    input_unique = new_w[new_w.sum(axis=1) == 0].index
    input_abd_missed: float = new_abd.loc[input_unique].sum().sum()
    input_abd_missed_prop: float = input_abd_missed / new_abd.sum().sum()
    logger.info(
        f"{len(input_unique)} features unique to input, could not be "
        f"matched any entries in signature W matrix.")
    logger.info(
        f"These unique features represent {input_abd_missed:.2f} weight of"
        f" a total {new_abd.sum().sum():.2f} ({input_abd_missed_prop:.2%}).")

    # Loss of W weight
    w_total: float = w.sum().sum()
    new_w_total: float = new_w.sum().sum()
    logger.info(
        f"Sum of weights in new W matrix is {new_w_total:.2f}; "
        f"Original signature matrix sum is {w_total:.2f}. "
        f"{1 - (new_w_total / w_total):.2%} of signature W matrix weight lost "
        f"due to feature mismatch.")

    # Transform to signature weights
    es: pd.DataFrame = nmf_transform(new_abd=new_abd,
                                     w_prime=new_w)
    decomp: Decomposition = Decomposition(
        w=new_w, h=es.T, feature_mapping=mapping,
        parameters=NMFParameters(
            x=new_abd,
            rank=new_w.shape[1],
            seed='Reapply'
        ),
    )
    decomp.colors = colors
    return decomp


def reapply(y: Union[str, pd.DataFrame],
            model: Union[str, Signatures] = "5es",
            hard_mapping: Optional[Union[str, pd.DataFrame]] = None,
            separator: str = "\t",
            output_dir: Optional[str] = None,
            **kwargs
            ) -> 'denovo.Decomposition':
    """Load and transform abundances to an existing model.

    The new data must be annotated against the same taxonomy the model uses.
    Currently for the 5 ES models this is GTDB r207. Feature names
    will be automatically matched between the abundance table and model where
    possible, (see :func:`match_genera`). Most of the work is done in
    :func:`transform_table`, this mostly provides convenience of allowing
    parameters to be paths or DataFrames, or to specify models as string or
    object.

    :param y: Feature matrix to transform. Can be a string giving 
        path, or a DataFrame.
    :param model: Model to use. Can be a Signature object, or the name
        of one of the provded Signature objects. Currently this is '5es' for the
        5ES model of Frioux et al.
        (2023, https://doi.org/10.1016/j.chom.2023.05.024).
    :param hard_mapping: Define matchups between feature identifiers in
        y and those in model W matrix. These will be used in preference of any
        automated matches. Should be a table with index being y matrix
        identifier, and first column the model W identifier. Can be either a
        path, or DataFrame.
    :param separator: Separator to use when reading and writing matrices.
    :param output_dir: Directory to write results to. Directory will be created
        if it does not exist. Pass None for no output to disk.
    :param **kwargs: Passed to the Signature validate_input and match_feature
        functions.
    """

    # Load files if required
    abundance_df: pd.DataFrame
    if isinstance(y, str):
        abundance_df = pd.read_csv(y, sep=separator, index_col=0)
    else:
        abundance_df = y

    # Load model
    model_obj: Signatures
    if isinstance(model, str):
        # This is either a known model identifier, or a location for model
        # Check for known models
        from cvanmf import models
        if model.lower() == "5es":
            logger.info("Loading 5ES model")
            model_obj = models.five_es()
        else:
            raise CVANMFException(f"Model '{model}' not recognised.")
    else:
        model_obj = model

    # Load hard mapping
    mapping_dict: Dict[str, str]
    if hard_mapping is not None:
        if isinstance(hard_mapping, str):
            mapping_dict = pd.read_csv(hard_mapping, sep=separator).to_dict()
        else:
            mapping_dict = hard_mapping.to_dict()
    else:
        mapping_dict = {}

    res: 'denovo.Decomposition' = model_obj.reapply(
        y=abundance_df,
        hard_mapping=mapping_dict,
        **kwargs
    )

    if output_dir is not None:
        res.save(out_dir=pathlib.Path(output_dir), delim=separator)

    return res


@click.command()
@click.option("-i",
              "--input",
              required=True,
              type=click.Path(exists=True, dir_okay=False),
              help="""New feature matrix, with features on rows and samples on 
              columns.""")
@click.option("-m",
              "--model",
              required=False,
              # TODO: Define the list of models elsewhere
              type=click.Choice(['5es']),
              default="5es",
              help="""Name of the model to reapply.""")
@click.option("-h",
              "--hard_mapping",
              required=False,
              type=click.Path(exists=True, dir_okay=False),
              help="""Mapping between features in input table and model W 
              matrix. Provide as a csv, with first column features in input 
              table, second column the name in model W to map to.""")
@click.option("--rollup/--no-rollup",
              default=True,
              help="""Only used when genera are features. 
              Genera in abundance table which do not match the model W matrix, 
              add their abundance to a family level entry if one exists.""")
@click.option("-s", "--separator",
              default="\t", type=str,
              help="""Separator used in input and output files.""")
@click.option("-o", "--output_dir",
              required=True, type=click.Path(file_okay=False),
              help="""Directory to write output to.""")
def cli(input: str,
        model: str,
        hard_mapping: str,
        rollup: bool,
        separator: str,
        output_dir: str) -> None:
    """Command line interface to fit new data to an existing NMF Signatures
    model. The new data must use the same features as the model,
    though there can be some difference (features in now data not in model
    and vice versa). Currently this is GTDB r207 for the 5 Enterosignatures
    model.
    
    For more on Enterosignatures see:

    *  Frioux et al. 2023 (doi:10.1016/j.chom.2023.05.024)
    *  https://enterosignatures.quadram.ac.uk
    """

    if model is None:
        model = '5es'

    # Use transform function
    res: 'Decomposition' = reapply(
        y=input,
        model=model,
        hard_mapping=hard_mapping,
        rollup=rollup,
        separator=separator,
        output_dir=output_dir
    )


from __future__ import annotations

import itertools
import logging
import math
from collections import Counter, namedtuple
from typing import Union, Tuple, Iterable, List, Set, Literal, Optional, Dict, \
    Callable, Any

import networkx as nx
import numpy as np
import pandas as pd
import plotnine
from scipy.stats import kruskal
from sklearn.metrics.pairwise import cosine_similarity

from cvanmf.denovo import Decomposition
from cvanmf.reapply import match_identical, _reapply_model
from cvanmf.stability import Comparable, align_signatures, match_signatures, \
    get_signatures_from_comparable


# Alias for types which hold signatures that can be compared


def combine_signatures(
        signatures: Iterable[Comparable],
        x: pd.DataFrame = None,
        merge_threshold: float = 0.9,
        split_threshold: float = 0.9,
        prune_low_support: bool = True,
        low_support_threshold: Union[int, float] = 0.2,
        low_support_alpha: float = 0.05
) -> Decomposition:
    """Combine signatures into a non-redundant set."""

    # 1. Cosine similarity between all
    logging.info("Concatenating signatures from models")
    aligned_sigs: List[pd.DataFrame] = align_signatures(signatures)
    cat_signatures: pd.DataFrame = pd.concat(
        aligned_sigs,
        axis=1
    )
    merged_sigs: pd.DataFrame = cat_signatures
    # Keep identifying and merging cliques until no cliques |c| > 1 remain
    while True:
        cos: pd.DataFrame = cosine_similarity(merged_sigs.T)

        # 2. Maximal clique detection
        # This might benefit from being replaced with some near-clique methods
        # Could replace with some clustering method instead if this seems too
        # crude
        # Apply threshold to get adjacency matrix, convert to graph, B-K to find
        # cliques
        logging.info("Identifying cliques in similarity graph")
        adj: pd.DataFrame = cos > merge_threshold
        g = nx.from_numpy_array(adj)
        cliques = list(nx.find_cliques(g))
        logging.info("Found %s cliques, merging to single signatures",
                     len(cliques))

        # If all cliques are a single node, terminate merging
        if max(len(x) for x in cliques) < 2:
            break

        # 3. Merging
        # Make a new dataframe with the merged signatures
        # After making, remove all deleted signatures
        # Concatenate on merged signatures
        merged_sigs_i: pd.DataFrame = pd.concat(
            [merged_sigs.iloc[:, c].mean(axis=1) for c in cliques],
            axis=1
        )
        merged_sigs = pd.concat(
            [merged_sigs_i,
             merged_sigs.iloc[:,
             [x for x in range(merged_sigs.shape[1])
              if x not in list(itertools.chain.from_iterable(cliques))]
             ]],
            axis=1
        )

    # 4a. Fit each signature S_a to S_{-a} to and see if good similarity
    # (meaning it can be described well as a mix as some other signatures)
    # and can probably be removed
    logging.info("Pruning signatures which are adequately described as a mix of"
                 "other signatures")
    sig_remove: List[int] = []
    for i in range(merged_sigs.shape[1]):
        retain_sigs: List[int] = [x for x in range(merged_sigs.shape[1])
                                  if x not in sig_remove and x != i]
        reduced_sigs: pd.DataFrame = merged_sigs.iloc[:, retain_sigs]
        reduced_model: Decomposition = _reapply_model(
            y=merged_sigs.iloc[:, i].to_frame(),
            w=reduced_sigs,
            input_validation=lambda x: x,
            feature_match=match_identical,
            colors=None
        )
        if reduced_model.model_fit.iloc[0] > split_threshold:
            sig_remove.append(i)
    logging.info("Removed %s signatures which are linear combinations"
                 "of others", len(sig_remove))
    merged_sigs = merged_sigs.iloc[:,
                  [x for x in range(merged_sigs.shape[1])
                   if x not in sig_remove]]

    if not prune_low_support:
        return merged_sigs

    # 5. Remove signatures with low support which do not significantly
    # change model fit when removed
    # Determine number of models which contained something similar to
    # this signature, using the merge threshold
    support: pd.Series = pd.concat(
        [pd.Series(
            (cosine_similarity(merged_sigs.T, x.T) > merge_threshold
             ).any(axis=1)
        )
            for x in aligned_sigs],
        axis=1
    ).sum(axis=1)
    # support is now a series with the number of models which contain a
    # signature similar to this
    req_support: int  # If signature in this or more models, is well supported
    match low_support_threshold:
        case int():
            req_support = low_support_threshold
        case float():
            req_support = math.ceil(len(aligned_sigs) * low_support_threshold)
        case _:
            raise TypeError('low_support_threshold should be int or float')
    req_support = max(1, min(req_support, len(aligned_sigs)))
    low_support_sigs: pd.DataFrame = support[support < req_support]
    # Test the effect of removing the low support signature from a model
    # containing only the well supported signatures and this one
    good_sigs: pd.DataFrame = merged_sigs.iloc[:,
                              [x for x in support.index
                               if x not in low_support_sigs.index]]
    reject_low: List[int] = []
    for poor_sig, support in low_support_sigs.items():
        i_sigs: pd.DataFrame = pd.concat(
            [merged_sigs.iloc[:, poor_sig],
             good_sigs],
            axis=1,
        )
        # Rename to avoid undesired behaviour with dropping column by index
        i_sigs.columns = [f's{i}' for i in i_sigs.columns]
        loo_modelfit: List[pd.Series] = [
            _reapply_model(y=x,
                           w=i_sigs.drop(columns=[i]),
                           colors=None,
                           input_validation=lambda x: x,
                           feature_match=match_identical).model_fit
            for i in i_sigs.columns
        ]
        h, p = kruskal(loo_modelfit[0],
                       pd.concat(loo_modelfit[1:]))
        if (loo_modelfit[0].mean() > pd.concat(loo_modelfit[1:]).mean() and
                p <= low_support_alpha):
            # Reject this signature
            reject_low.append(poor_sig)
    merged_sigs = merged_sigs.iloc[:, [x for x in range(merged_sigs.shape[1])
                                       if x not in reject_low]]

    return merged_sigs


class Signature(pd.Series):
    """One signature which is being combined."""

    def __init__(self, model: Optional[Model] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.model = model

    @property
    def model(self) -> Optional[Model]:
        return self.__model

    @model.setter
    def model(self, model: Optional[Model] = None) -> None:
        self.__model: Model = model

    @staticmethod
    def from_comparable(c: Comparable) -> List[Signature]:
        return [Signature(data=col) for _, col
                in get_signatures_from_comparable(c).T.iterrows()]

    @staticmethod
    def mds(signatures: Union[Iterable[Signature], pd.DataFrame],
            **kwargs) -> pd.DataFrame:
        """Perform MDS ordination of a set of signatures.

        Positions signatures in an n-dimensional space using cosine distance.
        Using the sklearn MDS implementation, so all arguments to the MDS
        constructor can be passed in kwargs. Generally useful ones are
        n_components for number of dimensions, metric for metric or non-metric
        MDS.

        :param signatures: Iterable of signatures, or a DataFrame of signatures
        :param n: Dimensions to use in NMDS
        :returns: DataFrame with signatures on rows and dimensions coordinates
        on columns.
        """
        from sklearn.manifold import MDS

        # Make a DataFrame is input is not already one
        sig_df: pd.DataFrame
        if not isinstance(signatures, pd.DataFrame):
            # Convert signatures to a list so we can guarantee ordered access
            # There's an outside chance we might be passed a list
            signatures = list(signatures)
            sig_df = pd.concat(signatures, axis=1)
        else:
            sig_df = signatures

        # Make cosine distance as input
        cos_dist: np.ndarray = 1.0 - cosine_similarity(sig_df.T)

        mds: MDS = MDS(dissimilarity="precomputed", **kwargs)
        transformed: np.ndarray = mds.fit_transform(cos_dist)
        transformed_df: pd.DataFrame = pd.DataFrame(
            transformed,
            index=sig_df.columns,
            columns=[f'dim_{i}' for i in range(1, mds.n_components + 1)]
        )

        if not isinstance(signatures, pd.DataFrame):
            # Attempt to attached cohort and model id as metadata where set
            transformed_df['cohort'] = [
                'None' if x.model.cohort is None else x.model.cohort.name
                for x in signatures
            ]
            transformed_df['model'] = [id(x.model) for x in signatures]
            transformed_df['model_rank'] = [len(x.model.signatures)
                                            for x in signatures]
        return transformed_df

    def __repr__(self) -> str:
        return f'Signature[id={id(self)}]'

    def __hash__(self) -> int:
        """Uses object memory address as hash value."""
        return id(self)


class Model:
    """One decomposition which is being combined."""

    def __init__(self,
                 signatures: Optional[Iterable[Signature]] = None,
                 cohort: Optional[Cohort] = None):
        self.signatures = signatures
        self.__cohort: Cohort = cohort

    @property
    def signatures(self) -> List[Signature]:
        return self.__signatures

    @signatures.setter
    def signatures(self, signatures: Iterable[Signature]) -> None:
        self.__signatures = list() if signatures is None else list(signatures)

    @property
    def cohort(self) -> Cohort:
        return self.__cohort

    @cohort.setter
    def cohort(self, cohort: Cohort) -> None:
        self.__cohort = cohort

    def add_signatures(self, signatures: Iterable[Signature]) -> Model:
        """Add signatures and set them to refer to this model."""

        for x in signatures:
            x.model = self
        self.signatures += list(signatures)
        return self

    @staticmethod
    def from_comparable(comparable: Comparable) -> Model:
        sigs: List[Signature] = Signature.from_comparable(comparable)
        model: Model = Model()
        model.add_signatures(sigs)
        return model


    def __repr__(self) -> str:
        return (f'Model[nsigs={len(self.signatures)}, '
                f'cohort='
                f'{"None" if self.cohort is None else self.cohort.name}]')


class Cohort:

    def __init__(self,
                 name: str,
                 models: Optional[List[Model]] = None,
                 x: Optional[pd.DataFrame] = None) -> None:
        self.name = name
        self.models = models
        self.x = x

    @property
    def name(self) -> str:
        return self.__name if self.__name is not None else str(id(self))

    @name.setter
    def name(self, name: str) -> None:
        self.__name = name

    @property
    def models(self) -> List[Model]:
        return self.__models

    @models.setter
    def models(self, models: Optional[List[Model]]) -> None:
        self.__models = [] if models is None else models

    @property
    def x(self) -> Optional[pd.DataFrame]:
        return self.__x

    @x.setter
    def x(self, x: Optional[pd.DataFrame]) -> None:
        self.__x: pd.DataFrame = x

    @property
    def signatures(self) -> List[Signature]:
        return list(itertools.chain.from_iterable(
            x.signatures for x in self.models))

    @staticmethod
    def from_comparables(
            comparables: Iterable[Comparable],
            name: str = "",
            x: Optional[pd.DataFrame] = None
    ) -> Cohort:
        models: List[Model] = [Model.from_comparable(x) for x in comparables]
        cohort: Cohort = Cohort(name=name, models=models, x=x)
        for m in cohort.models:
            m.cohort = cohort
        return cohort

    def add_models(self, models: Iterable[Model]) -> Cohort:
        for m in models:
            m.cohort = self
        self.models += list(models)
        return self

    def __repr__(self):
        return (f'Cohort[nmodels={len(self.models)}, '
                f'nsigs={len(self.signatures)}]')


class Cluster:

    def __init__(self,
                 signatures: Optional[Iterable[Signature]] = None,
                 label: Optional[str] = None):
        self.__signatures: List[Signature] = (
            list(signatures) if signatures is not None else [])
        self.label = label

    @property
    def signatures(self) -> List[Signature]:
        return self.__signatures

    @property
    def mean_signature(self) -> pd.Series:
        """Return the mean signature for this cluster."""

        return pd.concat(self.signatures, axis=1).mean(axis=1)

    @property
    def member_data(self) -> pd.DataFrame:
        """Data from all cohorts which any member signature originates from."""

        distinct_cohorts: Set[Cohort] = set(
            x.model.cohort for x in self.signatures
        )
        merged_df: pd.DataFrame = pd.concat([x.x for x in distinct_cohorts],
                                            axis=1)
        # If there are duplicate columns removed them but emit warning
        orig_shape: Tuple[int, int] = merged_df.shape
        merged_df = merged_df.loc[:, ~merged_df.columns.duplicated()]
        if orig_shape != merged_df.shape:
            logging.warning("Sample names duplicated between cohorts. If these "
                            "are distinct samples give them different names, "
                            "otherwise they will be treated as one sample "
                            "during signature combining.")
        return merged_df

    @property
    def label(self) -> str:
        """Label for this cluster in plots."""
        return f'Cluster_{id(self)}' if self.__label is None else self.__label

    @label.setter
    def label(self, label: str) -> None:
        """Setter for label."""
        self.__label: str = label

    def support(self,
                type: Literal["signature", "model", "cohort"] = "signature",
                samples: bool = False
                ) -> int:
        """The support for this cluster.

        We might want to place more confidence in signatures which appear more
        frequently, either having more similar signatures overall, appearing
        in more models, or appearing in more cohorts. This function provides
        either the number of signatures in the cluster, the number of models
        the cluster members originate from (likely to be the same as number of
        signatures), or the number of cohorts the signature is found in.
        Alternatively, we can express this as the total number of samples for
        the model or cohort case, for when some cohorts are small and we might
        want to weight those signatures from large cohorts.

        :param type: What to count. One of 'signature', 'model', 'cohort'
        :param samples: Count the number of samples for model or cohort
            rather than treating each as one.
        """

        if type == "signature" or type == "signatures":
            return len(self.signatures)
        elif type == "model" or type == "models":
            distinct_models = set(x.model for x in self.signatures)
            return sum(m.cohort.x.shape[1] for m in distinct_models) \
                if samples else len(distinct_models)
        elif type == "cohort" or type == "cohorts":
            distinct_cohorts = set(x.model.cohort for x in self.signatures)
            return sum(c.x.shape[1] for c in distinct_cohorts) \
                if samples else len(distinct_cohorts)
        else:
            raise ValueError("type must be one of 'signature', 'model', "
                             "'cohort'")

    @property
    def cohort_model_count(self) -> Counter:
        """Number of models from each cohort which contain member signatures."""

        return Counter((m.cohort.name, len(m.signatures)) for m in
                       set(s.model for s in self.signatures))

    def member_similarity(self,
                          utri: bool = False) -> pd.DataFrame:
        """Pairwise cosine similarity of the member signatures."""

        sig_df: pd.DataFrame = pd.concat(
            self.signatures,
            axis=1
        )
        sig_df.columns = range(sig_df.shape[1])
        sig_cs: pd.DataFrame = pd.DataFrame(
            cosine_similarity(sig_df.T),
            index=sig_df.columns,
            columns=sig_df.columns
        )
        sig_cs = sig_cs.stack(future_stack=True).reset_index()
        sig_cs.columns = ['a', 'b', 'cosine_similarity']
        sig_cs['cohort_a'] = sig_cs['a'].apply(
            lambda x: self.signatures[x].model.cohort.name
        )
        sig_cs['cohort_b'] = sig_cs['b'].apply(
            lambda x: self.signatures[x].model.cohort.name
        )
        sig_cs['within'] = sig_cs['cohort_a'] == sig_cs['cohort_b']
        # Remove diagonal
        sig_cs = sig_cs.loc[sig_cs['a'] != sig_cs['b'], :]
        sig_cs['pair'] = sig_cs.apply(
            lambda x: set([x['a'], x['b']]),
            axis=1
        )
        if utri:
            sig_cs = sig_cs.drop_duplicates(subset=['pair'])
        return sig_cs

    def plot_member_similarity(self,
                               split_cohort: bool = True) -> plotnine.ggplot:
        sim: pd.DataFrame = self.member_similarity()
        plt: plotnine.ggplot
        if split_cohort:
            plt = (
                    plotnine.ggplot(
                        sim,
                        mapping=plotnine.aes(
                            x="cohort_a",
                            y="cosine_similarity",
                            fill="within"
                        )
                    ) + plotnine.geom_boxplot()
            )
        else:
            plt = (
                    plotnine.ggplot(
                        sim.drop_duplicates(subset=['pair']),
                        mapping=plotnine.aes(
                            y="cosine_similarity"
                        )
                    ) +
                    plotnine.geom_boxplot()
            )
        return plt

    def member_feature_weights(self,
                               top_n: int = None,
                               unit_scale: bool = True) -> pd.DataFrame:
        """Weights of features in all member signatures of this cluster.

        This is returned in long form, intended to be used for plotting using
        plotnine."""

        feature_df: pd.DataFrame = pd.concat(
            self.signatures,
            axis=1
        )
        feature_df.columns = range(feature_df.shape[1])
        # Convert each to unit vectors if requested, to control differences in
        # scale between models with different ranks
        if unit_scale:
            feature_df = feature_df / np.linalg.norm(feature_df, axis=0)
        feature_df = feature_df.loc[
                     (feature_df
                      .mean(axis="columns")
                      .sort_values(ascending=False)
                      .index), :]
        feature_df = feature_df.loc[
                     feature_df.mean(axis="columns").sort_values(ascending=False).index, :
                     ]
        if top_n is None:
            top_n = feature_df.shape[0]
        feature_df = feature_df.iloc[:min(top_n, feature_df.shape[0]), :]
        feature_df_stack: pd.DataFrame = (
            feature_df
            .stack(future_stack=True)
            .reset_index()
        )
        feature_df_stack.columns = ['feature', 'signature', 'weight']
        feature_df_stack['feature'] = pd.Categorical(
            feature_df_stack['feature'],
            categories=reversed(feature_df.index),
            ordered=True
        )
        # Attach cohort name as columns
        feature_df_stack['cohort'] = feature_df_stack['signature'].apply(
            lambda x: self.signatures[x].model.cohort.name
        )
        return feature_df_stack

    def plot_feature_variance(
            self,
            top_n: Optional[int] = None,
            split_cohort: bool = False
    ):
        """Box plot for how much a feature weight varies."""
        # TODO: Needs scaling somehow?
        feature_df: pd.DataFrame = self.member_feature_weights(top_n=top_n)
        mapping: plotnine.aes = (
            plotnine.aes(y="weight", x="feature", fill="cohort")
            if split_cohort else
            plotnine.aes(y="weight", x="feature"))
        plt: plotnine.ggplot = (
                plotnine.ggplot(
                    feature_df,
                    mapping=mapping
                ) +
                plotnine.geom_boxplot() +
                plotnine.coord_flip()
        )
        return plt

    def plot_mds(self,
                 **kwargs) -> plotnine.ggplot:
        """Plot signatures in this cluster in a 2D space.

        Produce a 2D representation of member signatures of this cluster, with
        color indicting cohort and shape indicating rank. Can pass any arguments
        used by scikit-learn MDS constructor in kwargs.
        """
        df: pd.DataFrame = Signature.mds(self.signatures, **kwargs)
        return (
                plotnine.ggplot(df,
                                plotnine.aes(x='dim_1', y='dim_2', color='cohort',
                                             shape='factor(model_rank)')) +
                plotnine.geom_point() +
                plotnine.labs(x="", y="", shape="Rank", color="Cohort")
        )

    @staticmethod
    def as_dataframe(clusters: List[Cluster]) -> pd.DataFrame:
        """Concatenated table with the mean signatures for each cluster.

        :param clusters: Clusters to concatenate
        :returns: Concatenated DataFrame with each mean signature as a column
        """
        return pd.concat([x.mean_signature for x in clusters], axis=1)

    @staticmethod
    def cosine_similarity(clusters: List[Cluster]) -> pd.DataFrame:
        """Get the pairwise cosine similarity between clusters.

        :param clusters: Clusters to calculate similarity between
        :returns: DataFrame with pairwise cosine similarity
        """

        cat_df: pd.DataFrame = Cluster.as_dataframe(clusters)
        return cosine_similarity(cat_df.T)

    @staticmethod
    def merge(clusters: Iterable[Cluster]) -> Cluster:
        """Merge clusters.

        :param clusters: Cluster to merge.
        :returns: A new cluster containing the union of the signatures.
        """

        return Cluster(set(itertools.chain.from_iterable(
            x.signatures for x in clusters
        )))

    def __repr__(self) -> str:
        return f'Cluster[nsigs={len(self.signatures)}]'


class Combiner:

    def __init__(self,
                 cohorts: Iterable[Cohort]
                 ) -> None:
        self.__cohorts: List[Cohort] = list(cohorts)
        self.__clusters: List[Cluster] = [Cluster([y]) for y in
                                          itertools.chain.from_iterable((x.signatures for x in cohorts))
                                          ]
        self.__linear_combinations: List[
            Tuple[Signature, List[Signature]]] = []
        self.__removed_clusters: List[Cluster] = []

    @property
    def cohorts(self) -> List[Cohort]:
        return self.__cohorts

    @property
    def clusters(self) -> List[Cluster]:
        return self.__clusters

    @property
    def removed_clusters(self) -> List[Cluster]:
        return self.__removed_clusters

    def merge_similar(self,
                      cosine_threshold: float = 0.9,
                      density_threshold: float = 0.98
                      ) -> None:
        """Merge signatures which are highly similar based on cosine similarity.

        Signatures are grouped into clusters when they all share similarity
        greater than the specified threshold. A signature can end up in
        multiple clusters as a result of this grouping. Use
        refine_multimembers to force a singular membership after merging.

        :param cosine_threshold: Cosine similarity above which to consider
            samples similar.
        """

        loop_i: int = 1
        while True:
            cos_df: pd.DataFrame = Cluster.cosine_similarity(self.clusters)
            adj: np.ndarray = cos_df > cosine_threshold
            np.fill_diagonal(adj, False)
            g = nx.from_numpy_array(adj)

            cliques = list(nx.find_cliques(g))
            logging.info("Merge iteration %s: %s cliques found in %s clusters",
                         loop_i, len(cliques), len(self.clusters))

            # Terminate if no cliques |clique|>1 to merge
            if max(len(x) for x in cliques) < 2:
                break

            # Merge largest clique
            # TODO: Continue merging cliques after removing vertices which
            # have already been merged
            sorted_cliques: List[Set[int]] = list(reversed(sorted(
                cliques,
                key=len
            )))
            in_clusters: Set[int] = set()
            new_clusters: List[Cluster] = []
            for clique in sorted_cliques:
                remaining_members: Set[int] = set(clique).difference(in_clusters)
                if len(remaining_members) > 1:
                    new_cluster: Cluster = Cluster.merge(
                        [self.clusters[i] for i in remaining_members])
                    new_clusters.append(new_cluster)
                    in_clusters.update(remaining_members)
            # Remove those which were added to cliques
            # and add new clusters
            for rm in [self.clusters[i] for i in in_clusters]:
                self.__clusters.remove(rm)
            for c in new_clusters:
                self.__clusters.append(c)
            # # Merge clusters in cliques
            # self.__clusters = [
            #     Cluster.merge(
            #         [self.clusters[i] for i in clique]
            #     ) for clique in cliques
            # ]

    def identify_cohort_specific(
            self,
            cohort_proportion: float = 0.95
    ) -> List[Cluster]:
        """Find clusters which are consistent in one or more cohorts.

        A cluster may be unique to a cohort, being consistently recovered in
        data from that cohort but not in some others. These may appear to
        have poor support globally (considering all cohorts), but when
        there is strong consensus within a cohort it may be of interest to
        retain these clusters.
        """
        # Find maximum number of models for each cohort
        max_support: Dict[Tuple[str, int], int] = {}
        for cohort in self.cohorts:
            for rank, count in Counter(
                    len(m.signatures) for m in cohort.models).items():
                max_support[(cohort.name, rank)] = count
        cluster_df: pd.DataFrame = pd.concat(
            [pd.Series(c.cohort_model_count) for c in self.clusters],
            axis=1
        ).fillna(0)
        cluster_prop: pd.DataFrame = (
            (cluster_df.T / pd.Series(max_support))
            .stack(future_stack=True).stack(future_stack=True)
            .reset_index()
            .set_axis(['cluster_idx', 'rank', 'cohort', 'prop'], axis=1)
        )
        specific: pd.DataFrame = cluster_prop[
            cluster_prop['prop'] >= cohort_proportion]

        # For now just pull out those clusters and return
        return [self.clusters[i] for i in set(specific['cluster_idx'])]

    def remove_low_support(
            self,
            support_required: Union[float, int] = 0.2,
            support_type: Literal["signature", "model", "cohort"] = "model",
            support_samples: bool = False,
            signature_floor: int = 2,
            exempt_clusters: List[Cluster] = None,
            retain_alpha: float = 0.05,
            only_cohort_data: bool = False
    ) -> List[Cluster]:
        """Remove signatures do not appear frequently among models or cohorts.

        This looks for clusters with signatures which appear in a small number
        of models (or just a small number of signatures, or signatures
        from a small number of cohorts). Any signatures below the threshold
        for low support are retained only if removing them cause a worse
        model fit than removing the other signatures (the model fit for all
        the others pooled). This can be evaluated on the full data, or only on
        the data from the cohorts which member signatures are drawn from.

        :param support_required: Either proportion or absolute number of below
            which the signature will be considered to have low support.
        :param support_type: What to use to determine support; either the
            number of signatures, number of models cluster members appear in,
            or number of cohorts cluster members appear in.
        :param support_samples: For model or cohort, count the number of
            samples rather than each model or cohort as one. Useful when
            cohorts very uneven size.
        :param signature_floor: Remove any clusters with fewer than this
            number of member signatures.
        :param exempt_clusters: Clusters which will not be removed even if
            meeting low support criteria.
        :param retain_alpha: Signatures are retained if removing them
            has a significantly different impact on model fit compared
            to any good signature. Tested with a Kruskal-Wallis test using
            this parameter as a threshold. Set to 0 to reject all low support
            signatures.
        :param only_cohort_data: Only use data from the cohorts from which
            this cluster has support when evaluating the change in model fit
            from omitting this signature.
        """

        # Determine the maximum possible support
        support_total: int
        if support_type == 'signature' or support_type == 'signatures':
            if support_samples:
                logging.warning("support_samples not supported for signatures.")
            support_total = len(set(itertools.chain.from_iterable(
                c.signatures for c in self.clusters
            )))
        elif support_type == 'model' or support_type == 'models':
            support_total = (
                sum([x.x.shape[1] * len(x.models) for x in self.cohorts])
                if support_samples else
                sum(len(x.models) for x in self.cohorts)
            )
        elif support_type == 'cohort' or support_type == 'cohorts':
            support_total = (
                sum([x.x.shape[1] for x in self.cohorts])
                if support_samples else
                len(self.cohorts)
            )
        else:
            raise ValueError("support_type should be one of "
                             "'signature', 'model', or 'cohort'.")

        req_support: int
        match support_required:
            case int():
                req_support = support_required
            case float():
                req_support = math.ceil(support_total * support_required)
            case _:
                raise TypeError('support_required should be int or float')
        req_support = max(1, min(req_support, support_total))

        exempt_clusters = exempt_clusters if exempt_clusters is not None else []

        # Remove any clusters below the floor
        floored: List[Cluster] = [
            y for y in self.clusters if len(y.signatures) < signature_floor]
        for y in floored:
            self.clusters.remove(y)
            self.__removed_clusters.append(y)

        # Identify remaining low support clusters
        low_support_clusters: List[Cluster] = [
            x for x in self.clusters
            if x.support(type=support_type, samples=support_samples)
               < req_support
               and
               x not in exempt_clusters
        ]
        other_clusters: List[Cluster] = [
            x for x in self.clusters if x not in low_support_clusters
        ]

        # Does removing this signature have a bigger impact than removing
        # any of the non-low support signatures?
        good_df: pd.DataFrame = Cluster.as_dataframe(other_clusters)
        for i_low in low_support_clusters:
            # If the alpha is 0 or lower, will never retain so skip calculation
            if retain_alpha <= 0:
                self.clusters.remove(i_low)
                self.__removed_clusters.append(i_low)
                continue
            i_sigs: pd.DataFrame = pd.concat(
                [Cluster.as_dataframe([i_low]), good_df],
                axis=1,
            )
            # Rename to avoid undesired behaviour with dropping column by index
            i_sigs.columns = [f's{i}' for i in range(i_sigs.shape[1])]
            # Get the input data to evaluate model fit on
            x: pd.DataFrame = (
                i_low.member_data if only_cohort_data else self.cohort_data
            )

            # Find model fit for all samples with each signature removed
            loo_modelfit: List[pd.Series] = [
                _reapply_model(y=x,
                               w=i_sigs.drop(columns=[i]),
                               colors=None,
                               input_validation=lambda x: x,
                               feature_match=match_identical).model_fit
                for i in i_sigs.columns
            ]

            # Compare the low support signature model fit to the others
            concat_mf: pd.Series = pd.concat(loo_modelfit[1:])
            h, p = kruskal(loo_modelfit[0],
                           concat_mf)
            # Only retain is the model fit is worse when removing this
            # signature than the others, and this is significantly different
            if not (loo_modelfit[0].mean() < concat_mf.mean() and
                    p <= retain_alpha):
                self.clusters.remove(i_low)
                self.__removed_clusters.append(i_low)

    @property
    def cohort_data(self) -> pd.DataFrame:
        """Combined data for all cohorts."""

        merged_df: pd.DataFrame = pd.concat([x.x for x in self.cohorts],
                                            axis=1)
        # If there are duplicate columns removed them but emit warning
        orig_shape: Tuple[int, int] = merged_df.shape
        merged_df = merged_df.loc[:, ~merged_df.columns.duplicated()]
        if orig_shape != merged_df.shape:
            logging.warning("Sample names duplicated between cohorts. If these "
                            "are distinct samples give them different names, "
                            "otherwise they will be treated as one sample "
                            "during signature combining.")
        return merged_df

    def remove_linear_combinations(
            self,
            cosine_threshold: float = 0.9,
            min_small_support_ratio: float = 0.5,
            support_type: Literal["signature", "model", "cohort"] = "model",
            support_samples: bool = False,
            do_removal: bool = True
    ) -> None:
        """Remove signatures which can be expressed as linear combinations.

        Some signatures might be a combination of multiple others, which
        commonly co-occur in some of the cohorts. We can remove these and
        keep only the constituent signatures. However, do not want to discard
        well supported large signatures as they are combination of smaller but
        less well supported signatures.

        :param cosine_threshold: Model fit threshold
        :param do_removal: Remove the signatures which are combinations. Set to
            False to return the identified signatures without removing.
        """

        removed: List[Tuple[Cluster, List[Cluster]]] = []
        for cluster in self.clusters:
            # Reject any cluster whose support is not sufficient
            c_support: int = cluster.support(type=support_type,
                                             samples=support_samples)
            accept_clusters: List[Cluster] = [
                c for c in self.clusters
                if (c is not cluster and
                    (c.support(type=support_type, samples=support_samples) /
                     c_support) >= min_small_support_ratio
                    )
            ]
            # If one or zero clusters acceptable, will never be usefully
            # describe as a mix so skip further tests
            if len(accept_clusters) < 2:
                continue

            # Make a matrix of all the other signatures
            # and perform NNLS to describe candidate signature as a mixture
            w: pd.DataFrame = Cluster.as_dataframe(accept_clusters)
            signature_mix: Decomposition = _reapply_model(
                y=cluster.mean_signature.to_frame(),
                w=w,
                input_validation=lambda x: x,
                feature_match=match_identical,
                colors=None
            )

            if signature_mix.model_fit.iloc[0] < cosine_threshold:
                continue
            # Take any signatures which are required to describe 99% of the
            # total weight as those which make up the one to be removed
            repr: pd.Series = signature_mix.representative_signatures(
                threshold=0.99).iloc[:, 0]
            repr_clusters: List[Cluster] = list(itertools.compress(
                accept_clusters,
                list(repr)
            ))
            removed.append((cluster, repr_clusters))

        if do_removal:
            for cluster, components in removed:
                self.clusters.remove(cluster)
                self.__linear_combinations.append((cluster, components))

        return removed

    def remove_linear_combinations_2(
            self,
            cosine_threshold: float = 0.9,
            min_small_support_ratio: float = 0.5,
            support_type: Literal["signature", "model", "cohort"] = "model",
            support_samples: bool = False,
            do_removal: bool = True
    ):
        """Linear combination removal with removal of smaller poorly supported
        signatures along with larger ones."""

        to_test: List[Cluster] = sorted(
            self.clusters,
            key=lambda x: x.support(type=support_type,
                                    samples=support_samples)
        )
        removed: Set[Cluster] = set()

        while len(to_test) > 0:
            # Start from highest support cluster
            cluster: Cluster = to_test[-1]
            remaining_clusters: List[Cluster] = [
                x for x in self.clusters
                if x not in removed and x is not cluster
            ]
            c_support: int = cluster.support(
                type=support_type, samples=support_samples
            )
            threshold_met: List[bool] = [
                (c.support(type=support_type, samples=support_samples) /
                 c_support) > min_small_support_ratio
                for c in remaining_clusters
            ]

            # Make a matrix of all other remaining signatures and performing
            # NNLS to describe candidate signature as a mixture
            w: pd.DataFrame = Cluster.as_dataframe(remaining_clusters)
            signature_mix: Decomposition = _reapply_model(
                y=cluster.mean_signature.to_frame(),
                w=w,
                input_validation=lambda x: x,
                feature_match=match_identical,
                colors=None
            )

            if signature_mix.model_fit.iloc[0] < cosine_threshold:
                # This is not a mixture of any other signatures
                to_test.remove(cluster)
                continue

            # This is a mixture of some of the other signatures
            repr: pd.Series = signature_mix.representative_signatures(
                threshold=0.98).iloc[:, 0]
            repr_clusters: List[Cluster] = list(itertools.compress(
                remaining_clusters,
                list(repr)
            ))
            repr_threshold: List[bool] = list(itertools.compress(
                threshold_met,
                list(repr)
            ))
            # If all the signatures which contribute to this one fall above
            # the relative support threshold, remove this signature
            if all(repr_threshold):
                removed.update({cluster})
                self.__linear_combinations.append(
                    (cluster, repr_clusters)
                )
                to_test.remove(cluster)
            # Remove any signatures which fall below support threshold,
            # but do not remove this signature form the candidate pool
            else:
                for below in itertools.compress(repr_clusters,
                                                (not x for x in repr_threshold)):
                    removed.update({below})
                    if below in to_test:
                        to_test.remove(below)

        if do_removal:
            for x in removed:
                self.clusters.remove(x)

        return removed

    def plot_feature_variance(
            self,
            split_cohort: bool = True,
            top_n: Optional[int] = 20,
            label_fn: Callable[[List[str]], List[str]] = None,
            unit_scale: bool = True
    ) -> plotnine.ggplot:
        """Plot feature variance for all signatures."""
        feature_df: pd.DataFrame = pd.concat([
            (x.member_feature_weights(top_n=top_n)
             .assign(cluster=x.label))
            for i, x in enumerate(self.clusters)
        ])
        mapping: Dict[str, str] = dict(
            y="weight",
            x="feature"
        )
        if split_cohort:
            mapping['fill'] = "cohort"
        if label_fn is not None:
            feature_df['feature'] = label_fn(feature_df['feature'])
        plt: plotnine.ggplot = (
                plotnine.ggplot(
                    feature_df,
                    mapping=plotnine.aes(**mapping)
                ) +
                plotnine.geom_boxplot(outlier_size=0.2) +
                plotnine.coord_flip()
        )
        if split_cohort:
            plt = plt + plotnine.facet_wrap("cluster",
                                            scales="free_y",
                                            ncol=len(self.clusters))
        else:
            plt = plt + plotnine.facet_wrap(facets="cluster", scales="free_y")
        return plt

    def plot_member_similarity(self,
                               split_cohort: bool = True):
        sim: pd.DataFrame = pd.concat([
            (x.member_similarity()
             .assign(cluster=x.label))
            for i, x in enumerate(self.clusters)
        ])
        plt: plotnine.ggplot
        if split_cohort:
            plt = (
                    plotnine.ggplot(
                        sim,
                        mapping=plotnine.aes(
                            x="cohort_a",
                            y="cosine_similarity",
                            fill="within"
                        )
                    ) +
                    plotnine.geom_boxplot() +
                    plotnine.facet_wrap("cluster") +
                    plotnine.labs(x="Cohort", y="Cosine Similarity")
            )
        else:
            plt = (
                    plotnine.ggplot(
                        sim.drop_duplicates(subset=['pair']),
                        mapping=plotnine.aes(
                            x="cluster",
                            y="cosine_similarity"
                        )
                    ) +
                    plotnine.geom_boxplot() +
                    plotnine.labs(x="Cluster", y="Cosine Similarity")
            )
        return plt

    def plot_mds(
            self,
            hull: bool = True,
            **kwargs
    ) -> plotnine.ggplot:
        """Plot all clusters member signatures in reduced dimensions."""

        sig_list: List[Signature] = list(itertools.chain.from_iterable(
            x.signatures for x in self.clusters))
        signatures: pd.DataFrame = pd.concat(
            sig_list + [x.mean_signature for x in self.clusters], axis=1)
        df: pd.DataFrame = Signature.mds(signatures, **kwargs)

        # Attach metadata
        df = df.assign(cohort="Mean", model_rank=0, cluster=None)
        num_centroids: int = len(self.clusters)
        # For non-centroid signatures
        df['cohort'] = [
            x.model.cohort.name for x in sig_list] + (["Mean"] * num_centroids)
        df['model_rank'] = [
            len(x.model.signatures) for x in sig_list] + ([0] * num_centroids)
        df['cluster'] = list(
            itertools.chain.from_iterable(
                [x.label] * len(x.signatures) for x in self.clusters
            )) + list(x.label for x in self.clusters)
        # Label if is a mean signature
        df['is_mean'] = (
                np.array(range(df.shape[0])) < (df.shape[0] - num_centroids))
        plt: plotnine.ggplot = (
                plotnine.ggplot(
                    df,
                    plotnine.aes(x="dim_1", y="dim_2", color="cohort",
                                 shape="factor(model_rank)")) +
                plotnine.geom_point(size=0.5) +

                plotnine.geom_text(
                    plotnine.aes(label="cluster"),
                    data=df[~df['is_mean']],
                    size=6,
                    color="black"
                )
        )
        if hull:
            plt = plt + plotnine.stat_hull(
                plotnine.aes(group="cluster"), color="black")
        return plt

    def label_by_match(self,
                       external_model: Comparable) -> pd.DataFrame:
        """Label clusters by their similarity to an existing model."""
        match: pd.DataFrame = match_signatures(
            Cluster.as_dataframe(self.clusters),
            external_model
        )
        for match_row in match.loc[~match['a'].isna(), :].iterrows():
            match_row = match_row[1]
            cluster_idx: Optional[int] = match_row['a']
            other_lbl: Optional[str] = match_row['b']
            other_cos: Optional[float] = match_row['pair_cosine']
            if other_lbl is not None:
                self.clusters[cluster_idx].label = (
                    f'{other_lbl} ({other_cos:.2f})')
            else:
                self.clusters[cluster_idx].label = (
                    f'Cluster {cluster_idx}'
                )
        return match


def example_cohort_structure() -> Dict[str, pd.DataFrame]:
    """Make 5 cohorts of data with a complex structure.

    This set of cohorts is suitable to evaluate whether we are capturing the
    different kind of situations we expect during signature combining. These
    are:
    * highly similar signatures which should be merged
    * signatures which are linear combinations of others which should be removed
    * low support signatures (present in few cohorts) which can be
    *   *   uninformative, in which case removed
    *   *   informative cohort specific, in which case retained

    To this end, we make 5 cohorts based on the 5 ES model signatures.
    1.  Contains all 5 ES
    2.  Doesn't contain ES_Bifi
    3.  Contains an extra signature IS_1 which is informative
    4.  Completely shuffled, does not fit at all
    5.  Doesn't contain ES_Bact or ES_Esch
    6.  Merged ES_Firm and ES_Prev

    Each is returned as a named tuple containing name, h, w, x
    """
    from cvanmf import models

    CohortTuple = namedtuple("CohortTuple", "name w h x")
    # C1 - All ES
    es_w: pd.DataFrame = models.five_es().w
    c1_w: pd.DataFrame = es_w.copy()
    c1_h: pd.DataFrame = pd.DataFrame(
        np.random.uniform(low=0, high=1, size=(c1_w.shape[1], 100)),
        index=c1_w.columns
    )
    c1_h: pd.DataFrame = c1_h / c1_h.sum()
    c1_x: pd.DataFrame = c1_w.dot(c1_h)
    c1 = CohortTuple("All", c1_w, c1_h, c1_x / c1_x.sum())

    # C2 - No ES_Bifi
    c2_w: pd.DataFrame = es_w.drop(columns=['ES_Bifi'])
    c2_h: pd.DataFrame = pd.DataFrame(
        np.random.uniform(low=0, high=1, size=(c2_w.shape[1], 200)),
        index=c2_w.columns
    )
    c2_h: pd.DataFrame = c2_h / c2_h.sum()
    c2_x: pd.DataFrame = c2_w.dot(c2_h)
    c2 = CohortTuple("No_Bifi", c2_w, c2_h, c2_x / c2_x.sum())

    # C3 - Extra signature
    c3_w: pd.DataFrame = es_w.copy()
    c3_w.loc[:, 'IS_1'] = es_w['ES_Firm'].sample(frac=1).values
    c3_h: pd.DataFrame = pd.DataFrame(
        np.random.uniform(low=0, high=1, size=(c3_w.shape[1], 150)),
        index=c3_w.columns
    )
    c3_h: pd.DataFrame = c3_h / c3_h.sum()
    c3_x: pd.DataFrame = c3_w.dot(c3_h)
    c3 = CohortTuple("IS_1", c3_w, c3_h, c3_x / c3_x.sum())

    # C4 - Randomly shuffled, which should make irrelevant signatures
    c4_w: pd.DataFrame = es_w.copy()
    c4_h: pd.DataFrame = pd.DataFrame(
        np.random.uniform(low=0, high=1, size=(c4_w.shape[1], 100)),
        index=c4_w.columns
    )
    c4_h: pd.DataFrame = c4_h / c4_h.sum()
    c4_x: pd.DataFrame = c4_w.dot(c4_h)
    for i in range(c4_x.shape[1]):
        c4_x.iloc[:, i] = c4_x.iloc[:, i].sample(frac=1)
    c4 = CohortTuple("Random", c4_w, c4_h, c4_w / c4_w.sum())

    # C5 - No Bifi or Esch
    c5_w: pd.DataFrame = es_w.drop(columns=['ES_Bact', 'ES_Esch'])
    c5_h: pd.DataFrame = pd.DataFrame(
        np.random.uniform(low=0, high=1, size=(c5_w.shape[1], 200)),
        index=c5_w.columns
    )
    c5_h: pd.DataFrame = c5_h / c5_h.sum()
    c5_x: pd.DataFrame = c5_w.dot(c5_h)
    c5 = CohortTuple("No_Bifi_Esch", c5_w, c5_h, c5_x / c5_x.sum())

    # C6 - Bact and Prev merged
    c6_w: pd.DataFrame = es_w.copy()
    c6_w['ES_Prev+Bact'] = c6_w['ES_Prev'] + c6_w['ES_Bact']
    c6_w = c6_w.drop(columns=['ES_Bact', 'ES_Prev'])
    c6_h: pd.DataFrame = pd.DataFrame(
        np.random.uniform(low=0, high=1, size=(c6_w.shape[1], 200)),
        index=c6_w.columns
    )
    c6_h: pd.DataFrame = c6_h / c6_h.sum()
    c6_x: pd.DataFrame = c6_w.dot(c6_h)
    c6 = CohortTuple("Merge_Prev_Bact", c6_w, c6_h, c6_x / c6_x.sum())

    return {x.name: x for x in [c1, c2, c3, c4, c5, c6]}


def split_dataframe_to_cohorts(
        x: pd.DataFrame,
        cohort_labels: pd.Series,
        min_size: int = 0
) -> Dict[Any, pd.DataFrame]:
    """Split a DataFrame into multiple, based on the provided cohort labels."""

    cohort_count: pd.Series = cohort_labels.groupby(cohort_labels).count()
    cohort_include: pd.Series = cohort_count.loc[cohort_count >= min_size]
    labels: pd.Series = cohort_labels.loc[
        cohort_labels.isin(cohort_include.index)]
    return {
        lbl: x.loc[:, cohort_labels.loc[cohort_labels == lbl].index]
        for lbl in cohort_include.index
    }




"""Load existing Enterosignature models."""
import logging
from importlib.resources import files
from typing import NamedTuple, List, Optional, Dict, Union

import pandas as pd

logger: logging.Logger = logging.getLogger(__name__)

FIVE_ES_COLORS = {
    "ES_Bact": "#E69F00",
    "ES_Firm": "#023e8a",
    "ES_Prev": "#D55E00",
    "ES_Bifi": "#009E73",
    "ES_Esch": "#483838"
}


class Signatures(NamedTuple):
    """Definition of an existing signature model.

    This provides the definition of existing signatures required to reapply
    the signature model to new data. Where Decomposition stores the input and
    H matrix, these are not necessary for transforming new data. Rather, we
    only need the W matrix, the colors associated with each signature (for
    consistency of representation), and the preprocessing steps (to match
    features in the new data with those in the W matrix)."""
    w: pd.DataFrame
    """Feature weights (W matrix) for this model."""
    colors: Union[List[str], Dict[str, str]]
    """Color for each signature in the model."""
    feature_match: 'FeatureMatch'
    """Function to map features in new data to those in the model W matrix."""
    input_validation: 'InputValidation' = lambda x: x
    """Function to validate and potentially transform input table. Defaults
    to identity function"""
    citation: Optional[str] = None
    """Citation when using this model."""

    def reapply(self,
                y: pd.DataFrame,
                **kwargs) -> 'Decomposition':
        """Transform new data using this signature model.

        :param y: New data of same type as the existing model.
        """
        from cvanmf import reapply
        return reapply._reapply_model(
            y=y,
            **(self._asdict() | kwargs)
        )


def five_es() -> Signatures:
    """The 5 Enterosignature model of Frioux et al.
    (2023, https://doi.org/10.1016/j.chom.2023.05.024). A summary of this model
    can also be found on the website https://enterosignatures.quadram.ac.uk.
    The `reapply` method for this model will normalise (total-sum-scale) input
    data after applying filters to match model format, so data provided does
    not need to be normalised.

    :return: 5 Enterosignature model
    :type: Signatures
    """

    w: pd.DataFrame = pd.read_csv(
        str(files("cvanmf.data").joinpath("ES5_W.tsv")),
        sep="\t",
        index_col=0
    )
    citation: str = (
        "Frioux, C. et al. Enterosignatures define common bacterial guilds in "
        "the human gut microbiome. Cell Host & Microbe 31, 1111-1125.e6 ("
        "2023). https://doi.org/10.1016/j.chom.2023.05.024")
    logger.warning("If you use the 5ES model please cite %s",
                   citation)
    from cvanmf import reapply
    return Signatures(w=w,
                      colors=FIVE_ES_COLORS,
                      feature_match=reapply.match_genera,
                      input_validation=reapply.validate_genus_table,
                      citation=(
                          "Frioux, C. et al. Enterosignatures define common "
                          "bacterial guilds in the human gut microbiome. "
                          "Cell Host & Microbe 31, 1111-1125.e6 (2023)."
                          " https://doi.org/10.1016/j.chom.2023.05.024")
                      )
